@@ -9,7 +9,7 @@
 '''command to display statistics about repository history'''
 
 from mercurial.i18n import _
-from mercurial import patch, cmdutil, util, templater, commands
+from mercurial import patch, cmdutil, scmutil, util, templater, commands
 import os
 import time, datetime
 
@@ -24,7 +24,7 @@ def maketemplater(ui, repo, tmpl):
 
 def changedlines(ui, repo, ctx1, ctx2, fns):
     added, removed = 0, 0
-    fmatch = cmdutil.matchfiles(repo, fns)
+    fmatch = scmutil.matchfiles(repo, fns)
     diff = ''.join(patch.diff(repo, ctx1.node(), ctx2.node(), fmatch))
     for l in diff.split('\n'):
         if l.startswith("+") and not l.startswith("+++ "):
@@ -54,13 +54,13 @@ def countrate(ui, repo, amap, *pats, **opts):
     if opts.get('date'):
         df = util.matchdate(opts['date'])
 
-    m = cmdutil.match(repo, pats, opts)
+    m = scmutil.match(repo[None], pats, opts)
     def prep(ctx, fns):
         rev = ctx.rev()
         if df and not df(ctx.date()[0]): # doesn't match date format
             return
 
-        key = getkey(ctx)
+        key = getkey(ctx).strip()
         key = amap.get(key, key) # alias remap
         if opts.get('changesets'):
             rate[key] = (rate.get(key, (0,))[0] + 1, 0)
@@ -129,8 +129,14 @@ def churn(ui, repo, *pats, **opts):
         aliases = repo.wjoin('.hgchurn')
     if aliases:
         for l in open(aliases, "r"):
-            alias, actual = l.split('=' in l and '=' or None, 1)
-            amap[alias.strip()] = actual.strip()
+            try:
+                alias, actual = l.split('=' in l and '=' or None, 1)
+                amap[alias.strip()] = actual.strip()
+            except ValueError:
+                l = l.strip()
+                if l:
+                    ui.warn(_("skipping malformed alias: %s\n" % l))
+                continue
 
     rate = countrate(ui, repo, amap, *pats, **opts).items()
     if not rate:
@@ -143,13 +149,14 @@ def churn(ui, repo, *pats, **opts):
     maxcount = float(max(sum(v) for k, v in rate)) or 1.0
     maxname = max(len(k) for k, v in rate)
 
-    ttywidth = util.termwidth()
+    ttywidth = ui.termwidth()
     ui.debug("assuming %i character terminal\n" % ttywidth)
     width = ttywidth - maxname - 2 - 2 - 2
 
     if opts.get('diffstat'):
         width -= 15
-        def format(name, (added, removed)):
+        def format(name, diffstat):
+            added, removed = diffstat
             return "%s %15s %s%s\n" % (pad(name, maxname),
                                        '+%d/-%d' % (added, removed),
                                        ui.label('+' * charnum(added),
