@@ -3,7 +3,7 @@
 # Copyright 2006 Vadim Gelfer <vadim.gelfer@gmail.com>
 #
 # This software may be used and distributed according to the terms of the
-# GNU General Public License version 2, incorporated herein by reference.
+# GNU General Public License version 2 or any later version.
 
 '''hooks for sending email notifications at commit/push time
 
@@ -21,6 +21,8 @@ this::
   incoming.notify = python:hgext.notify.hook
   # batch emails when many changesets incoming at one time
   changegroup.notify = python:hgext.notify.hook
+  # batch emails when many changesets outgoing at one time (client side)
+  outgoing.notify = python:hgext.notify.hook
 
   [notify]
   # config items go here
@@ -37,7 +39,8 @@ Optional configuration items::
   style = ...            # style file to use when formatting email
   template = ...         # template to use when formatting email
   incoming = ...         # template to use when run as incoming hook
-  changegroup = ...      # template when run as changegroup hook
+  outgoing = ...         # template to use when run as outgoing hook
+  changegroup = ...      # template to use when run as changegroup hook
   maxdiff = 300          # max lines of diffs to include (0=none, -1=all)
   maxsubject = 67        # truncate subject line longer than this
   diffstat = True        # add a diffstat before the diff content
@@ -134,7 +137,7 @@ class notifier(object):
             c = path.find('/')
             if c == -1:
                 break
-            path = path[c+1:]
+            path = path[c + 1:]
             count -= 1
         return path
 
@@ -215,8 +218,8 @@ class notifier(object):
                 s = ctx.description().lstrip().split('\n', 1)[0].rstrip()
                 subject = '%s: %s' % (self.root, s)
         maxsubject = int(self.ui.config('notify', 'maxsubject', 67))
-        if maxsubject and len(subject) > maxsubject:
-            subject = subject[:maxsubject-3] + '...'
+        if maxsubject:
+            subject = util.ellipsis(subject, maxsubject)
         msg['Subject'] = mail.headencode(self.ui, subject,
                                          self.charsets, self.test)
 
@@ -249,7 +252,7 @@ class notifier(object):
     def diff(self, ctx, ref=None):
 
         maxdiff = int(self.ui.config('notify', 'maxdiff', 300))
-        prev = ctx.parents()[0].node()
+        prev = ctx.p1().node()
         ref = ref and ref.node() or ctx.node()
         chunks = patch.diff(self.repo, prev, ref, opts=patch.diffopts(self.ui))
         difflines = ''.join(chunks).splitlines()
@@ -290,7 +293,7 @@ def hook(ui, repo, hooktype, node=None, source=None, **kwargs):
     ui.pushbuffer()
     data = ''
     count = 0
-    if hooktype == 'changegroup':
+    if hooktype == 'changegroup' or hooktype == 'outgoing':
         start, end = ctx.rev(), len(repo)
         for rev in xrange(start, end):
             if n.node(repo[rev]):
